/*
    Copyright (C) 2003 Nikolas Zimmermann <wildfox@kde.org>
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    aint with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <iostream>

#include "Glyph.h"
#include "BezierPathAgg.h"
#include "GlyphTracerAgg.h"

using namespace T2P;

int traceMoveto(FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathAgg *bpath = static_cast<BezierPathAgg *>(glyph->modifiableBezierPath());

	bpath->storage.move_to(to->x * affine[0] + to->y * affine[2],
							to->x * affine[1] + to->y * affine[3]);

	return 0;
}

int traceLineto(FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathAgg *bpath = static_cast<BezierPathAgg *>(glyph->modifiableBezierPath());

	bpath->storage.line_to(to->x * affine[0] + to->y * affine[2],
							to->x * affine[1] + to->y * affine[3]);

	return 0;
}

int traceConicBezier(FT_Vector *control, FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathAgg *bpath = static_cast<BezierPathAgg *>(glyph->modifiableBezierPath());

	double cx = control->x * affine[0] + control->y * affine[2];
	double cy = control->x * affine[1] + control->y * affine[3];
	double ex3 = to->x * affine[0] + to->y * affine[2];
	double ey3 = to->x * affine[1] + to->y * affine[3];
	double sx3, sy3;
	bpath->storage.vertex(bpath->storage.total_vertices() - 1, &sx3, &sy3);

	bpath->storage.curve4(cx - (cx - sx3) / 3, cy - (cy - sy3) / 3, cx + (ex3 - cx) / 3, cy + (ey3 - cy) / 3, ex3, ey3);

	return 0;
}

int traceCubicBezier(FT_Vector *control1, FT_Vector *control2, FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathAgg *bpath = static_cast<BezierPathAgg *>(glyph->modifiableBezierPath());

	double x1 = control1->x * affine[0] + control1->y * affine[2];
	double y1 = control1->x * affine[1] + control1->y * affine[3];
	double x2 = control2->x * affine[0] + control2->y * affine[2];
	double y2 = control2->x * affine[1] + control2->y * affine[3];
	double x3 = to->x * affine[0] + to->y * affine[2];
	double y3 = to->x * affine[1] + to->y * affine[3];

	bpath->storage.curve4(x1, y1, x2, y2, x3, y3);

	return 0;
}

GlyphTracerAgg::GlyphTracerAgg() : GlyphTracer()
{
	setMoveto(*traceMoveto);
	setLineto(*traceLineto);
	setConicBezier(*traceConicBezier);
	setCubicBezier(*traceCubicBezier);
}

GlyphTracerAgg::~GlyphTracerAgg()
{
}

void GlyphTracerAgg::correctGlyph(GlyphAffinePair *glyphAffine)
{

	// Take bezier path belonging to glyph (Note: that one is _UNMODIFIABLE_, once calculated)
	const BezierPathAgg *path = static_cast<const BezierPathAgg *>(glyphAffine->glyph()->bezierPath());

	BezierPathAgg *transformatedPath = static_cast<BezierPathAgg *>(allocBezierPath(0));
	double *affine = glyphAffine->affine();
	transformatedPath->copy_from(path->storage, affine);
	glyphAffine->setTransformatedPath(transformatedPath);

//	bpath->m_transform = agg::affine_matrix(affine[0], affine[1], affine[2], affine[3], affine[4], affine[5]); 
}

BezierPath *GlyphTracerAgg::allocBezierPath(int)
{
	return new BezierPathAgg();
}	  

void GlyphTracerAgg::closePath(Glyph *glyph)
{
	BezierPathAgg *bpath = static_cast<BezierPathAgg *>(glyph->modifiableBezierPath());
	bpath->storage.close_polygon();
}

// vim:ts=4:noet
