/*
    Copyright (C) 2003 Nikolas Zimmermann <wildfox@kde.org>
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    aint with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <iostream>

#include "Glyph.h"
#include "BezierPathLibart.h"
#include "GlyphTracerLibart.h"

using namespace T2P;

int traceMoveto(FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathLibart *path = static_cast<BezierPathLibart *>(glyph->modifiableBezierPath());

	ArtPoint p;
	p.x = to->x * affine[0] + to->y * affine[2];
	p.y = to->x * affine[1] + to->y * affine[3];

	if(path->end == 0 ||
       p.x != path->path[path->end - 1].x3 ||
       p.y != path->path[path->end - 1].y3)
	{
		path->path[path->end].code = ART_MOVETO;
		path->path[path->end].x3 = p.x;
		path->path[path->end].y3 = p.y;
		path->end++;
	}

    return 0;
}

int traceLineto(FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathLibart *path = static_cast<BezierPathLibart *>(glyph->modifiableBezierPath());

	ArtPoint p;
	p.x = to->x * affine[0] + to->y * affine[2];
	p.y = to->x * affine[1] + to->y * affine[3];

	ArtBpath *last = &path->path[path->end - 1];

	if((p.x != last->x3) || (p.y != last->y3))
	{
		path->path[path->end].code = ART_LINETO;
		path->path[path->end].x3 = p.x;
		path->path[path->end].y3 = p.y;
		path->end++;
	}

	return 0;
}

int traceConicBezier(FT_Vector *control, FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathLibart *path = static_cast<BezierPathLibart *>(glyph->modifiableBezierPath());

	if(!(path->end > 0))
		return -1;

	ArtBpath *s = &path->path[path->end - 1];
	ArtBpath *e = &path->path[path->end];

	e->code = ART_CURVETO;

	ArtPoint c;
	c.x = control->x * affine[0] + control->y * affine[2];
	c.y = control->x * affine[1] + control->y * affine[3];
	e->x3 = to->x * affine[0] + to->y * affine[2];
	e->y3 = to->x * affine[1] + to->y * affine[3];

	path->path[path->end].x1 = c.x - (c.x - s->x3) / 3;
	path->path[path->end].y1 = c.y - (c.y - s->y3) / 3;
	path->path[path->end].x2 = c.x + (e->x3 - c.x) / 3;
	path->path[path->end].y2 = c.y + (e->y3 - c.y) / 3;

	path->end++;

	return 0;
}

int traceCubicBezier(FT_Vector *control1, FT_Vector *control2, FT_Vector *to, void *obj)
{
	Glyph *glyph = reinterpret_cast<Glyph *>(obj);
	double *affine = glyph->affine();
	BezierPathLibart *path = static_cast<BezierPathLibart *>(glyph->modifiableBezierPath());

	path->path[path->end].code = ART_CURVETO;
	path->path[path->end].x1 = control1->x * affine[0] + control1->y * affine[2];
	path->path[path->end].y1 = control1->x * affine[1] + control1->y * affine[3];
	path->path[path->end].x2 = control2->x * affine[0] + control2->y * affine[2];
	path->path[path->end].y2 = control2->x * affine[1] + control2->y * affine[3];
	path->path[path->end].x3 = to->x * affine[0] + to->y * affine[2];
	path->path[path->end].y3 = to->x * affine[1] + to->y * affine[3];

	path->end++;

	return 0;
}

GlyphTracerLibart::GlyphTracerLibart() : GlyphTracer()
{
	setMoveto(*traceMoveto);
	setLineto(*traceLineto);
	setConicBezier(*traceConicBezier);
	setCubicBezier(*traceCubicBezier);
}

GlyphTracerLibart::~GlyphTracerLibart()
{
}

void GlyphTracerLibart::correctGlyph(GlyphAffinePair *glyphAffine)
{
	// Take bezier path belonging to glyph (Note: that one is _UNMODIFIABLE_, once calculated)
	const BezierPathLibart *path = static_cast<const BezierPathLibart *>(glyphAffine->glyph()->bezierPath());

	// Create a new empty path with the same size
	BezierPathLibart *transformatedPath = static_cast<BezierPathLibart *>(allocBezierPath(0));
	transformatedPath->path = art_bpath_affine_transform(path->path, glyphAffine->affine());
	glyphAffine->setTransformatedPath(transformatedPath);
}

BezierPath *GlyphTracerLibart::allocBezierPath(int size)
{
	BezierPathLibart *bpath = new BezierPathLibart();
	if(size != 0)
		bpath->path = art_new(ArtBpath, size);

	return bpath;
}	  

void GlyphTracerLibart::closePath(Glyph *glyph)
{
	BezierPathLibart *path = static_cast<BezierPathLibart *>(glyph->modifiableBezierPath());
	path->path[path->end].code = ART_END;
}

// vim:ts=4:noet
